"""
groupme-bot

A simple bot builder for GroupMe
"""
from .bot import Bot, Context
from .router import Router
from .callback import Callback
from .attachment import ImageAttachment, LocationAttachment, SplitAttachment, EmojiAttachment, MentionsAttachment

__version__ = "0.1.7"
__author__ = "Branden Colen"
__all__ = [
    "Bot", "Router", "Callback", "Context",
    "ImageAttachment", "LocationAttachment",
    "SplitAttachment", "EmojiAttachment",
    "MentionsAttachment"
]



class Callback:
    __slots__ = ['attachments', 'avatar_url', 'created_at', 'group_id', 'id',
                 'name', 'sender_id', 'sender_type', 'source_guid', 'system',
                 'text', 'user_id']

    def __init__(self, flask_response: dict):
        """
        The callback provided by GroupMe. When any message is sent to the group, GroupMe sends a callback to the
        bots endpoint. This object is generated from the callback JSON to allow for easy use by the bot handlers.

        :param dict flask_response: A dict with the callback json extracted from the flask response.
        """
        self.attachments = None
        self.avatar_url = None
        self.created_at = None
        self.group_id = None
        self.id = None
        self.name = None
        self.sender_id = None
        self.sender_type = None
        self.source_guid = None
        self.system = None
        self.text = None
        self.user_id = None
        for key, value in flask_response.items():
            self.__setattr__(key, value)



import atexit
import logging
from typing import List

from waitress import serve
from flask import request, Flask, jsonify
from apscheduler.schedulers.background import BackgroundScheduler, BaseScheduler

from .bot import Bot, Context
from .callback import Callback


class RouteExistsError(Exception):
    pass


class Router(object):
    RESERVED_ROUTES = ('/', '/_health')

    def __init__(self, flask_app: Flask = None, scheduler: BaseScheduler = None):
        """
        The Router is the primary object used to run the GroupMe Bot. Multiple Bots can be handled in one single
        router object. Each bot is assigned an endpoint path and requests to that endpoint will be handled by the
        associated bot.

        :param Flask flask_app: (optional) Default to Flask(__name__)
        :param BaseScheduler scheduler: (optional) Default to BackgroundScheduler(daemon=True)
        """
        self._app = flask_app if flask_app else Flask(__name__)
        self._scheduler = scheduler if scheduler else BackgroundScheduler(daemon=True)

        # build an memory store for bot objects
        self._bots = {}

        # define a logger
        self._logger = logging.Logger('Router')
        self._logger.setLevel(logging.INFO)

        # add reserved routes for summary and health check
        self._app.add_url_rule(
            '/',
            'summary',
            lambda: jsonify({'endpoints': self.endpoints, 'jobs': self.jobs}),
            methods=['GET']
        )
        self._app.add_url_rule(
            '/_health',
            '_health',
            lambda: 'OK',
            methods=['GET']
        )

    @property
    def bot_store(self) -> dict[str, Bot]:
        """
        The current store of routes and the associated bot.
        :return dict[str, Bot]: key=the endpoint route, value=the bot object associated
        """
        return self._bots

    @property
    def flask_app(self) -> Flask:
        """
        The Flask app objects being used for the router.
        :return Flask:
        """
        return self._app

    @property
    def scheduler(self) -> BaseScheduler:
        """
        The BaseScheduler object being used for cron jobs
        :return BaseScheduler:
        """
        return self._scheduler

    @property
    def logger(self) -> logging.Logger:
        """
        The Logger being used
        :return logging.Logger:
        """
        return self._logger

    @property
    def endpoints(self) -> dict[str, str]:
        """
        A summary of all endpoints within the Router
        :return dict[str, str]: key=the endpoint path, value=the string repr of the associated Bot
        """
        endpoint_summary = {}
        for path, bot in self._bots.items():
            endpoint_summary[path] = str(bot)
        return endpoint_summary

    @property
    def jobs(self) -> List[str]:
        """
        A list of all currently running scheduler jobs
        :return List[str]: A list of the string repr of each running job
        """
        job_summary = []
        jobs = self._scheduler.get_jobs()
        if jobs:
            for job in jobs:
                job_summary.append(str(job))
        return job_summary

    def add_bot(self, bot: Bot, callback_path: str) -> None:
        """Add a new bot to be run by the Router

        :param bot: The bot to be run
        :param callback_path: The callback path for which the bot can be accesses
        :return:
        """
        # perform validity checks
        if callback_path in self.RESERVED_ROUTES:
            raise RouteExistsError(
                'Cannot use one of the reserved routes. Reserved routes are: ' + str(self.RESERVED_ROUTES))

        if callback_path in self._bots:
            raise RouteExistsError(
                "Callback path `{}` is already in use by a separate bot. Must use a new route for each bot."
                .format(callback_path))

        # store the bot for later access
        self._bots[callback_path] = bot

        # add the flask route
        self._app.add_url_rule(
            callback_path,
            callback_path.replace("/", "") + '-callback-handler',
            self._handle_callback,
            methods=['POST']
        )

        # add scheduler jobs
        for job in bot.cron_jobs:
            self._scheduler.add_job(
                job['func'],
                job['trigger'],
                args=(Context(bot, Callback({})),),
                **job['kwargs']
            )

    def run(self, host: str = "0.0.0.0", port: int = 8000, debug: bool = False) -> None:
        """Run the Router with all bot handlers and scheduled jobs.

        :param str host: The host address to run the Router.
        :param int port: The port to run the Router.
        :param bool debug: (default: False) Whether or not to run in debug mode.
        """
        # start the cron scheduler and setup a shutdown on exit
        self._scheduler.start()
        atexit.register(lambda: self._scheduler.shutdown(wait=False))

        # run the flask app
        if debug:
            self._app.run(host=host, port=port, debug=debug)
        else:
            self._logger.setLevel(logging.ERROR)
            serve(self._app, host=host, port=port)

    def _handle_callback(self) -> str:
        request_data = request.get_json(silent=True)
        ctx = Context(self._bots[request.path], Callback(request_data))
        bot = self._bots[request.path]
        try:
            bot.handle_callback(ctx)
            self._logger.info({'status': 'SUCCESS', 'bot': str(bot), 'request': request_data})
            return 'Success'
        except Exception as e:
            self._logger.error({'status': 'ERROR', 'bot': str(bot), 'request': request_data}, exc_info=e)
            return str(e)



from typing import List, Union


class Attachment:
    def __init__(self, **kwargs):
        self._kwargs = kwargs

    def to_dict(self):
        out = {}
        for key, value in self._kwargs.items():
            if key in ('charmap', 'loci', 'user_ids'):
                out[key] = value if value else []
            else:
                out[key] = str(value)
        return out


class ImageAttachment(Attachment):
    def __init__(self, image_url: str):
        Attachment.__init__(self, type='image', url=image_url)


class LocationAttachment(Attachment):
    def __init__(self, name: str, lat: Union[str, float, int], lng: Union[str, float, int]):
        Attachment.__init__(self, type='location', name=name, lat=lat, lng=lng)


class SplitAttachment(Attachment):
    def __init__(self, token: str):
        Attachment.__init__(self, type='split', token=token)


class EmojiAttachment(Attachment):
    def __init__(self, placeholder: str, charmap: List[List[Union[int, int]]]):
        Attachment.__init__(self, type='emoji', placeholder=placeholder, charmap=charmap)


class MentionsAttachment(Attachment):
    def __init__(self, loci: List[List[int]], user_ids: List[int]):
        Attachment.__init__(self, type='mentions', loci=loci, user_ids=user_ids)



from unittest import TestCase

from .bot import Bot
from .router import Router, RouteExistsError


class TestRouter(TestCase):

    def test_add_bot(self):
        router = Router()

        with self.assertRaises(RouteExistsError):
            router.add_bot(Bot("", "", "", ""), '/_health')

        with self.assertRaises(RouteExistsError):
            router.add_bot(Bot("", "", "", ""), '/')

        router.add_bot(Bot("fake", "", "", ""), "/bot")
        self.assertEqual(list(router.bot_store.keys()), ['/bot'])
        self.assertIsInstance(router.bot_store['/bot'], Bot)



import os

from setuptools import setup

import groupme_bot


def read(file_name):
    return open(os.path.join(os.path.dirname(__file__), file_name)).read()


setup(
    name='groupme-bot',
    version=groupme_bot.__version__,
    description='A simple bot builder for GroupMe',
    url='https://github.com/brandenc40/groupme-bot',
    author=groupme_bot.__author__,
    author_email='brandencolen@anl.gov',
    license='MIT',
    packages=['groupme_bot'],
    long_description=read('README.md'),
    long_description_content_type="text/markdown",
    install_requires=[
        'flask',
        'waitress',
        'apscheduler',
        'requests'
    ],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Programming Language :: Python :: 3',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ],
)



from __future__ import annotations

import json
import re
from typing import Any, List, Callable

import requests

from .attachment import Attachment, MentionsAttachment
from .callback import Callback

BASE_URL = 'https://api.groupme.com/v3'
BOT_INDEX_URL = BASE_URL + '/bots'
BOT_POST_URL = BASE_URL + '/bots/post'
GROUPS_URL = BASE_URL + '/groups/'
GROUP_ME_IMAGES_URL = 'https://image.groupme.com/pictures'


class HandlerPatternExistsError(Exception):
    pass


class Context:
    __slots__ = ['bot', 'callback']

    def __init__(self, bot: Bot, callback: Callback):
        """
        Context provided to every handler/scheduled function run by the bot. This provides a clean object containing
        bot the bot in use and the Callback that was sent.

        :param Bot bot:
        :param Callback callback:
        """
        self.bot = bot
        self.callback = callback


class Bot(object):
    def __init__(self, bot_name: str, bot_id: str, groupme_api_token: str, group_id: str):
        """
        The Bot class represents a single bot that can contains multiple callback handlers and scheduled jobs.
        Bots are run using the Router class.

        :param bot_name: A unique name for this bot to be known as within this code. Does not impact the name
            displayed in the GroupMe app.
        :param bot_id: The Bot ID provided by GroupMe.
        :param groupme_api_token: The GroupMe API token for access to group details.
        :param group_id: The id of the GroupMe group in which the bot exists.
        """
        self.bot_name = bot_name
        self.bot_id = bot_id
        self.api_token = groupme_api_token
        self.group_id = group_id

        self._handler_functions = {}
        self._cron_jobs = []

    @property
    def cron_jobs(self) -> List[dict]:
        """
        All list of cron jobs associated with this bot.

        :return List[dict]:
        """
        return self._cron_jobs

    def __str__(self):
        return "%s: %d callback handlers, %d cron jobs" \
               % (self.bot_name, len(self._handler_functions), len(self._cron_jobs))

    def add_callback_handler(self, regex_pattern: str, func: Callable[[Context], Any]) -> None:
        """
        Registers a regex pattern as to a bot handler function. If the regex pattern
        is found in a message from a GroupMe user, the function will be called.

        :param regex_pattern: The pattern to search for in the message text
        :param Callable[[Context], Any] func: The function to be called when the pattern is matched
        """
        if regex_pattern in self._handler_functions:
            raise HandlerPatternExistsError("The pattern `%s` is already registered to a handler" % regex_pattern)
        self._handler_functions[regex_pattern] = func

    def add_cron_job(self, func: Callable[[Context], Any], **kwargs) -> None:
        """
        Registers a function to be run on set cron schedule.
        
        Uses APScheduler cron trigger. Details here: https://apscheduler.readthedocs.io/en/stable/modules/triggers/cron.html
        
        Available arguments:
            - year (int|str) – 4-digit year
            - month (int|str) – month (1-12)
            - day (int|str) – day of the (1-31)
            - week (int|str) – ISO week (1-53)
            - day_of_week (int|str) – number or name of weekday (0-6 or mon,tue,wed,thu,fri,sat,sun)
            - hour (int|str) – hour (0-23)
            - minute (int|str) – minute (0-59)
            - second (int|str) – second (0-59)
            - start_date (datetime|str) – earliest possible date/time to trigger on (inclusive)
            - end_date (datetime|str) – latest possible date/time to trigger on (inclusive)
            - timezone (datetime.tzinfo|str) – time zone to use for the date/time calculations (defaults to scheduler timezone)
            - jitter (int|None) – advance or delay the job execution by jitter seconds at most.

        :param Callable[[Context], Any] func: The function to be called when the cron trigger is triggered
        """
        self._cron_jobs.append({
            'func': func,
            'trigger': 'cron',
            'kwargs': kwargs
        })

    def handle_callback(self, ctx: Context) -> None:
        """
        The main method used to handle incoming messages. Callbacks will not be handled if they are messages that came
        from the bot itself to prevent infinite loops.

        :param Context ctx: The Context of the request containing the Callback and the Bot objects.
        """
        if ctx.callback.user_id == self.bot_id:
            return
        text = ctx.callback.text.lower().strip()
        for pattern, func in self._handler_functions.items():
            if re.search(pattern, text):
                func(ctx)

    def post_message(self, msg: str, attachments: List[Attachment] = None) -> requests.Response:
        """
        Posts a bot message to the group with optional attachments.

        :param str msg: The message to be sent
        :param List[Attachment] attachments: (optional) Attachments to send in the message
        :return requests.Response: The POST request response object
        """
        if attachments:
            attachments = [attachment.to_dict() for attachment in attachments]
        else:
            attachments = []
        data = {
            "bot_id": self.bot_id,
            "text": msg,
            "attachments": attachments
        }
        response = requests.post(BOT_POST_URL, data=json.dumps(data), headers={'Content-Type': 'application/json'})
        response.raise_for_status()
        return response

    def image_url_to_groupme_image_url(self, image_url: str) -> str:
        """Convert a normal image URL to a GroupMe image to allow for usage as an attachment

        :param str image_url: The URL for any image
        :return str: The URL for the converted GroupMe image
        """
        r = requests.get(image_url)
        r.raise_for_status()
        headers = {
            'X-Access-Token': self.api_token,
            'Content-Type': r.headers['Content-type'],
        }
        response = requests.post(GROUP_ME_IMAGES_URL, headers=headers, data=r.content)
        response.raise_for_status()
        out = response.json()
        return out['payload']['picture_url']

    def get_group_summary(self) -> dict:
        """
        Get a summary of the group from the GroupMe API

        :return dict:
        """
        out = requests.get(GROUPS_URL + self.group_id, params={'token': self.api_token})
        out.raise_for_status()
        return out.json()['response']

    def mention_all(self) -> None:
        """
        Mentions everybody in the group so they receive a notification
        """
        text = ''
        user_ids = []
        loci = []
        for member in self.get_group_summary()['members']:
            user_ids.append(member['user_id'])
            loci.append([len(text), len(member['nickname']) + 1])
            text += '@{} '.format(member['nickname'])
        self.post_message(text, [MentionsAttachment(loci=loci, user_ids=user_ids)])
